from .predefined import (aam_fastest_alternating_noise,
                         aam_fastest_alternating_bbox,
                         aam_best_performance_alternating_noise,
                         aam_best_performance_alternating_bbox,
                         aam_params_combinations_noise,
                         clm_params_combinations_noise, clm_basic_noise,
                         clm_basic_bbox)
from .io import download_ibug_bounding_boxes
